#!/usr/bin/env python
"""scriptworker exceptions."""

from typing import Any

from scriptworker.constants import STATUSES


class ScriptWorkerException(Exception):
    """The base exception in scriptworker.

    When raised inside of the run_loop loop, set the taskcluster task
    status to at least ``self.exit_code``.

    Attributes:
        exit_code (int): this is set to 5 (internal-error).

    """

    exit_code = STATUSES["internal-error"]


class ScriptWorkerRetryException(ScriptWorkerException):
    """Scriptworker retry error.

    Attributes:
        exit_code (int): this is set to 4 (resource-unavailable)

    """

    exit_code = STATUSES["resource-unavailable"]


class ScriptWorkerTaskException(ScriptWorkerException):
    """Scriptworker task error.

    To use::

        import sys
        try:
            ...
        except ScriptWorkerTaskException as exc:
            log.exception("log message")
            sys.exit(exc.exit_code)

    Attributes:
        exit_code (int): this is 1 by default (failure)

    """

    def __init__(self, *args: Any, exit_code: int = 1):
        """Initialize ScriptWorkerTaskException.

        Args:
            *args: These are passed on via super().
            exit_code (int, optional): The exit_code we should exit with when
                this exception is raised.  Defaults to 1 (failure).

        """
        self.exit_code = exit_code
        super(ScriptWorkerTaskException, self).__init__(*args)


class TaskVerificationError(ScriptWorkerTaskException):
    """Verification error on a Taskcluster task.

    Use it when your script fails to validate any input from the task definition

    """

    def __init__(self, msg: str):
        """Initialize TaskVerificationError.

        Args:
            msg (string): the error message

        """
        super().__init__(msg, exit_code=STATUSES["malformed-payload"])


class BaseDownloadError(ScriptWorkerTaskException):
    """Base class for DownloadError and Download404.

    Attributes:
        exit_code (int): this is set to 4 (resource-unavailable).

    """

    def __init__(self, msg: str):
        """Initialize Download404.

        Args:
            msg (string): the error message

        """
        super(BaseDownloadError, self).__init__(msg, exit_code=STATUSES["resource-unavailable"])


class Download404(BaseDownloadError):
    """404 in ``scriptworker.utils.download_file``.

    Attributes:
        exit_code (int): this is set to 4 (resource-unavailable).

    """


class DownloadError(BaseDownloadError):
    """Failure in ``scriptworker.utils.download_file``.

    Attributes:
        exit_code (int): this is set to 4 (resource-unavailable).

    """


class CoTError(ScriptWorkerTaskException, KeyError):
    """Failure in Chain of Trust verification.

    Attributes:
        exit_code (int): this is set to 3 (malformed-payload).

    """

    def __init__(self, msg: str):
        """Initialize CoTError.

        Args:
            msg (string): the error message

        """
        super(CoTError, self).__init__(msg, exit_code=STATUSES["malformed-payload"])


class ScriptWorkerEd25519Error(CoTError):
    """Scriptworker ed25519 error.

    Attributes:
        exit_code (int): this is set to 5 (internal-error).

    """

    def __init__(self, msg: str):
        """Initialize ScriptWorkerEd25519Error.

        Args:
            msg (string): the error message

        """
        super(ScriptWorkerEd25519Error, self).__init__(msg)


class ConfigError(ScriptWorkerException):
    """Invalid configuration provided to scriptworker.

    Attributes:
        exit_code (int): this is set to 5 (internal-error).

    """


class WorkerShutdownDuringTask(BaseException):
    """Task cancelled because worker is shutting down."""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Deal with the scriptworker version in semver format.

Copied from scriptharness.

However, since writing this I've discovered that setuptools and sphinx don't
accept all semver formatted versions.  It's not clear if this will go away.

When called as a script, this will update ../version.json with the appropriate
version info.

Attributes:
  __version__ (Tuple[int, int, int, str]): semver version - three integers and an
    optional string.
  __version_string__ (str): semver version in string format.

"""
from __future__ import absolute_import, division, print_function, unicode_literals

import json
import os
from typing import Optional, Tuple, Union, cast

ShortVerType = Tuple[int, int, int]
LongVerType = Tuple[int, int, int, str]


# get_version_string {{{1
def get_version_string(version: Union[ShortVerType, LongVerType]) -> str:
    """Translate a version tuple into a string.

    Specify the __version__ as a tuple for more precise comparisons, and
    translate it to __version_string__ for when that's needed.

    This function exists primarily for easier unit testing.

    Args:
      version (Tuple[int, int, int, str]): three ints and an optional string.

    Returns:
      version_string (str): the tuple translated into a string per PEP 440

    """
    version_len = len(version)
    if version_len == 3:
        version_string = "%d.%d.%d" % cast(ShortVerType, version)
    elif version_len == 4:
        version_string = "%d.%d.%d.%s" % cast(LongVerType, version)
    else:
        raise Exception("Version tuple is non-semver-compliant {} length!".format(version_len))
    return version_string


# 1}}}
# Semantic versioning 2.0.0  http://semver.org/
__version__ = (42, 1, 0)
__version_string__ = get_version_string(__version__)


# write_version {{{1
def write_version(name: Optional[str] = None, path: Optional[str] = None) -> None:
    """Write the version info to ../version.json, for setup.py.

    Args:
      name (Optional[str]): this is for the ``write_version(name=__name__)``
        below.  That's one way to both follow the
        ``if __name__ == '__main__':`` convention but also allow for full
        coverage without ignoring parts of the file.

      path (Optional[str]): the path to write the version json to.  Defaults
        to ../version.json
    """
    # Written like this for coverage purposes.
    # http://stackoverflow.com/questions/5850268/how-to-test-or-mock-if-name-main-contents/27084447#27084447
    if name in (None, "__main__"):
        path = path or os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))), "version.json")
        contents = {"version": __version__, "version_string": __version_string__}
        with open(path, "w") as filehandle:
            print(json.dumps(contents, sort_keys=True, indent=4, separators=(",", ":")), file=filehandle)


write_version(name=__name__)



#!/usr/bin/env python
"""ed25519 support for scriptworker.

Attributes:
    log (logging.Logger): the log object for the module

"""
import argparse
import base64
import functools
import logging
import sys
from binascii import Error as Base64Error

from cryptography.exceptions import InvalidSignature, UnsupportedAlgorithm
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey, Ed25519PublicKey

from scriptworker.constants import DEFAULT_CONFIG
from scriptworker.exceptions import ScriptWorkerEd25519Error, ScriptWorkerException
from scriptworker.utils import read_from_file

log = logging.getLogger(__name__)


def ed25519_private_key_from_string(string):
    """Create an ed25519 private key from ``string``, which is a seed.

    Args:
        string (str): the string to use as a seed.

    Returns:
        Ed25519PrivateKey: the private key

    """
    try:
        return Ed25519PrivateKey.from_private_bytes(base64.b64decode(string))
    except (UnsupportedAlgorithm, Base64Error) as exc:
        raise ScriptWorkerEd25519Error("Can't create Ed25519PrivateKey: {}!".format(str(exc)))


def ed25519_public_key_from_string(string):
    """Create an ed25519 public key from ``string``, which is a seed.

    Args:
        string (str): the string to use as a seed.

    Returns:
        Ed25519PublicKey: the public key

    """
    try:
        return Ed25519PublicKey.from_public_bytes(base64.b64decode(string))
    except (UnsupportedAlgorithm, Base64Error) as exc:
        raise ScriptWorkerEd25519Error("Can't create Ed25519PublicKey: {}!".format(str(exc)))


def _ed25519_key_from_file(fn, path):
    """Create an ed25519 key from the contents of ``path``.

    ``path`` is a filepath containing a base64-encoded ed25519 key seed.

    Args:
        fn (callable): the function to call with the contents from ``path``
        path (str): the file path to the base64-encoded key seed.

    Returns:
        obj: the appropriate key type from ``path``

    Raises:
        ScriptWorkerEd25519Error

    """
    try:
        return fn(read_from_file(path, exception=ScriptWorkerEd25519Error))
    except ScriptWorkerException as exc:
        raise ScriptWorkerEd25519Error("Failed calling {} for {}: {}!".format(fn, path, str(exc)))


ed25519_private_key_from_file = functools.partial(_ed25519_key_from_file, ed25519_private_key_from_string)
ed25519_public_key_from_file = functools.partial(_ed25519_key_from_file, ed25519_public_key_from_string)


def ed25519_private_key_to_string(key):
    """Convert an ed25519 private key to a base64-encoded string.

    Args:
        key (Ed25519PrivateKey): the key to write to the file.

    Returns:
        str: the key representation as a str

    """
    return base64.b64encode(
        key.private_bytes(encoding=serialization.Encoding.Raw, format=serialization.PrivateFormat.Raw, encryption_algorithm=serialization.NoEncryption()), None
    ).decode("utf-8")


def ed25519_public_key_to_string(key):
    """Convert an ed25519 public key to a base64-encoded string.

    Args:
        key (Ed25519PublicKey): the key to write to the file.

    Returns:
        str: the key representation as a str

    """
    return base64.b64encode(key.public_bytes(encoding=serialization.Encoding.Raw, format=serialization.PublicFormat.Raw), None).decode("utf-8")


def verify_ed25519_signature(public_key, contents, signature, message):
    """Verify that ``signature`` comes from ``public_key`` and ``contents``.

    Args:
        public_key (Ed25519PublicKey): the key to verify the signature
        contents (bytes): the contents that was signed
        signature (bytes): the signature to verify
        message (str): the error message to raise.

    Raises:
        ScriptWorkerEd25519Error: on failure

    """
    try:
        public_key.verify(signature, contents)
    except InvalidSignature as exc:
        raise ScriptWorkerEd25519Error(message % {"exc": str(exc)})


# verify_ed25519_signature_cmdln {{{1
def verify_ed25519_signature_cmdln(args=None, exception=SystemExit):
    """Verify an ed25519 signature from the command line.

    Args:
        args (list, optional): the commandline args to parse. If ``None``, use
            ``sys.argv[1:]``. Defaults to ``None``.
        exception (Exception, optional): the exception to raise on failure.
            Defaults to ``SystemExit``.

    """
    args = args or sys.argv[1:]
    parser = argparse.ArgumentParser(
        description="""Verify an ed25519 signature from the command line.

Given a file and its detached signature, verify that it has been signed with
a valid key. This key can be specified on the command line; otherwise we'll
default to ``config['ed25519_public_keys']``."""
    )
    parser.add_argument("--pubkey", help="path to a base64-encoded ed25519 pubkey, optional")
    parser.add_argument("file_path")
    parser.add_argument("sig_path")
    opts = parser.parse_args(args)
    log = logging.getLogger("scriptworker")
    log.setLevel(logging.DEBUG)
    logging.basicConfig()
    pubkeys = {}
    if opts.pubkey:
        pubkeys["cmdln"] = [read_from_file(opts.pubkey)]
    pubkeys.update(dict(DEFAULT_CONFIG["ed25519_public_keys"]))
    contents = read_from_file(opts.file_path, file_type="binary")
    signature = read_from_file(opts.sig_path, file_type="binary")
    for key_type, seeds in pubkeys.items():
        for seed in seeds:
            try:
                verify_ed25519_signature(ed25519_public_key_from_string(seed), contents, signature, "didn't work with {}".format(seed))
                log.info("Verified good with {} seed {} !".format(key_type, seed))
                sys.exit(0)
            except ScriptWorkerEd25519Error:
                pass
    raise exception("This is not a valid signature!")



#!/usr/bin/env python
"""Generic utils for scriptworker.

Attributes:
    log (logging.Logger): the log object for the module

"""
import asyncio
import functools
import hashlib
import json
import logging
import os
import random
import re
import shutil
import time
from copy import deepcopy
from typing import IO, TYPE_CHECKING, Any, Awaitable, Callable, Dict, Match, Optional, Sequence, Tuple, Type, Union, cast, overload
from urllib.parse import unquote, urlparse

import aiohttp
import arrow
import async_timeout
import yaml
from taskcluster.client import createTemporaryCredentials

from scriptworker.exceptions import Download404, DownloadError, ScriptWorkerException, ScriptWorkerRetryException, ScriptWorkerTaskException

if TYPE_CHECKING:
    # Avoid circular import
    from scriptworker.context import Context
else:
    Context = object


log = logging.getLogger(__name__)


# request {{{1
async def request(context, url, timeout=60, method="get", good=(200,), retry=tuple(range(500, 512)), return_type="text", **kwargs):
    """Async aiohttp request wrapper.

    Args:
        context (scriptworker.context.Context): the scriptworker context.
        url (str): the url to request
        timeout (int, optional): timeout after this many seconds. Default is 60.
        method (str, optional): The request method to use.  Default is 'get'.
        good (list, optional): the set of good status codes.  Default is (200, )
        retry (list, optional): the set of status codes that result in a retry.
            Default is tuple(range(500, 512)).
        return_type (str, optional): The type of value to return.  Takes
            'json' or 'text'; other values will return the response object.
            Default is text.
        **kwargs: the kwargs to send to the aiohttp request function.

    Returns:
        object: the response text() if return_type is 'text'; the response
            json() if return_type is 'json'; the aiohttp request response
            object otherwise.

    Raises:
        ScriptWorkerRetryException: if the status code is in the retry list.
        ScriptWorkerException: if the status code is not in the retry list or
            good list.

    """
    session = context.session
    loggable_url = get_loggable_url(url)
    async with async_timeout.timeout(timeout):
        log.debug("{} {}".format(method.upper(), loggable_url))
        async with session.request(method, url, **kwargs) as resp:
            log.debug("Status {}".format(resp.status))
            message = "Bad status {}".format(resp.status)
            if resp.status in retry:
                raise ScriptWorkerRetryException(message)
            if resp.status not in good:
                raise ScriptWorkerException(message)
            if return_type == "text":
                return await resp.text()
            elif return_type == "json":
                return await resp.json()
            else:
                return resp


# retry_request {{{1
async def retry_request(*args, retry_exceptions=(asyncio.TimeoutError, ScriptWorkerRetryException), retry_async_kwargs=None, **kwargs):
    """Retry the ``request`` function.

    Args:
        *args: the args to send to request() through retry_async().
        retry_exceptions (list, optional): the exceptions to retry on.
            Defaults to (ScriptWorkerRetryException, ).
        retry_async_kwargs (dict, optional): the kwargs for retry_async.
            If None, use {}.  Defaults to None.
        **kwargs: the kwargs to send to request() through retry_async().

    Returns:
        object: the value from request().

    """
    retry_async_kwargs = retry_async_kwargs or {}
    return await retry_async(request, retry_exceptions=retry_exceptions, args=args, kwargs=kwargs, **retry_async_kwargs)


# datestring_to_timestamp {{{1
def datestring_to_timestamp(datestring):
    """Create a timetamp from a taskcluster datestring.

    Args:
        datestring (str): the datestring to convert. isoformat, like
            "2016-04-16T03:46:24.958Z"

    Returns:
        int: the corresponding timestamp.

    """
    return arrow.get(datestring).int_timestamp


# to_unicode {{{1
def to_unicode(line: Union[str, bytes]) -> str:
    """Avoid ``b'line'`` type messages in the logs.

    Args:
        line (str): The bytecode or unicode string.

    Returns:
        str: the unicode-decoded string, if ``line`` was a bytecode string.
            Otherwise return ``line`` unmodified.

    """
    ret = cast(str, line)
    try:
        ret = cast(bytes, line).decode("utf-8")
    except (UnicodeDecodeError, AttributeError):
        pass
    return ret


# makedirs {{{1
def makedirs(path: str) -> None:
    """Equivalent to mkdir -p.

    Args:
        path (str): the path to mkdir -p

    Raises:
        ScriptWorkerException: if path exists already and the realpath is not a dir.

    """
    if path:
        if not os.path.exists(path):
            log.debug("makedirs({})".format(path))
            os.makedirs(path)
        else:
            realpath = os.path.realpath(path)
            if not os.path.isdir(realpath):
                raise ScriptWorkerException("makedirs: {} already exists and is not a directory!".format(path))


# rm {{{1
def rm(path):
    """Equivalent to rm -rf.

    Make sure ``path`` doesn't exist after this call.  If it's a dir,
    shutil.rmtree(); if it's a file, os.remove(); if it doesn't exist,
    ignore.

    Args:
        path (str): the path to nuke.

    """
    if path and os.path.exists(path):
        if os.path.isdir(path):
            shutil.rmtree(path)
        else:
            os.remove(path)


# cleanup {{{1
def cleanup(context):
    """Clean up the work_dir and artifact_dir between task runs, then recreate.

    Args:
        context (scriptworker.context.Context): the scriptworker context.

    """
    for name in "work_dir", "artifact_dir", "task_log_dir":
        path = context.config[name]
        if os.path.exists(path):
            log.debug("rm({})".format(path))
            rm(path)
        makedirs(path)


# calculate_sleep_time {{{1
def calculate_sleep_time(attempt, delay_factor=5.0, randomization_factor=0.5, max_delay=120):
    """Calculate the sleep time between retries, in seconds.

    Based off of `taskcluster.utils.calculateSleepTime`, but with kwargs instead
    of constant `delay_factor`/`randomization_factor`/`max_delay`.  The taskcluster
    function generally slept for less than a second, which didn't always get
    past server issues.

    Args:
        attempt (int): the retry attempt number
        delay_factor (float, optional): a multiplier for the delay time.  Defaults to 5.
        randomization_factor (float, optional): a randomization multiplier for the
            delay time.  Defaults to .5.
        max_delay (float, optional): the max delay to sleep.  Defaults to 120 (seconds).

    Returns:
        float: the time to sleep, in seconds.

    """
    if attempt <= 0:
        return 0

    # We subtract one to get exponents: 1, 2, 3, 4, 5, ..
    delay = float(2 ** (attempt - 1)) * float(delay_factor)
    # Apply randomization factor.  Only increase the delay here.
    delay = delay * (randomization_factor * random.random() + 1)
    # Always limit with a maximum delay
    return min(delay, max_delay)


# retry_async {{{1
async def retry_async(
    func: Callable[..., Awaitable[Any]],
    attempts: int = 5,
    sleeptime_callback: Callable[..., Any] = calculate_sleep_time,
    retry_exceptions: Union[Type[BaseException], Tuple[Type[BaseException], ...]] = Exception,
    args: Sequence[Any] = (),
    kwargs: Optional[Dict[str, Any]] = None,
    sleeptime_kwargs: Optional[Dict[str, Any]] = None,
    log_exceptions: Optional[bool] = False,
) -> Any:
    """Retry ``func``, where ``func`` is an awaitable.

    Args:
        func (function): an awaitable function.
        attempts (int, optional): the number of attempts to make.  Default is 5.
        sleeptime_callback (function, optional): the function to use to determine
            how long to sleep after each attempt.  Defaults to ``calculateSleepTime``.
        retry_exceptions (list or exception, optional): the exception(s) to retry on.
            Defaults to ``Exception``.
        args (list, optional): the args to pass to ``func``.  Defaults to ()
        kwargs (dict, optional): the kwargs to pass to ``func``.  Defaults to
            {}.
        sleeptime_kwargs (dict, optional): the kwargs to pass to ``sleeptime_callback``.
            If None, use {}.  Defaults to None.

    Returns:
        object: the value from a successful ``function`` call

    Raises:
        Exception: the exception from a failed ``function`` call, either outside
            of the retry_exceptions, or one of those if we pass the max
            ``attempts``.

    """
    kwargs = kwargs or {}
    attempt = 1
    while True:
        try:
            return await func(*args, **kwargs)
        except retry_exceptions as exc:
            if log_exceptions:
                log.warning(f"retry_async exception:\n{type(exc)} {exc}")
            attempt += 1
            _check_number_of_attempts(attempt, attempts, func, "retry_async")
            await asyncio.sleep(_define_sleep_time(sleeptime_kwargs, sleeptime_callback, attempt, func, "retry_async"))


def retry_sync(func, attempts=5, sleeptime_callback=calculate_sleep_time, retry_exceptions=Exception, args=(), kwargs=None, sleeptime_kwargs=None):
    """Retry ``func``, where ``func`` is a regular function.

    Please favor ``retry_async`` whenever possible.

    Args:
        func (function): a function.
        attempts (int, optional): the number of attempts to make.  Default is 5.
        sleeptime_callback (function, optional): the function to use to determine
            how long to sleep after each attempt.  Defaults to ``calculateSleepTime``.
        retry_exceptions (list or exception, optional): the exception(s) to retry on.
            Defaults to ``Exception``.
        args (list, optional): the args to pass to ``func``.  Defaults to ()
        kwargs (dict, optional): the kwargs to pass to ``func``.  Defaults to
            {}.
        sleeptime_kwargs (dict, optional): the kwargs to pass to ``sleeptime_callback``.
            If None, use {}.  Defaults to None.

    Returns:
        object: the value from a successful ``function`` call

    Raises:
        Exception: the exception from a failed ``function`` call, either outside
            of the retry_exceptions, or one of those if we pass the max
            ``attempts``.

    """
    kwargs = kwargs or {}
    attempt = 1
    while True:
        try:
            return func(*args, **kwargs)
        except retry_exceptions:
            attempt += 1
            _check_number_of_attempts(attempt, attempts, func, "retry_sync")
            time.sleep(_define_sleep_time(sleeptime_kwargs, sleeptime_callback, attempt, func, "retry_sync"))


def _check_number_of_attempts(attempt: int, attempts: int, func: Callable[..., Any], retry_function_name: str) -> None:
    if attempt > attempts:
        log.warning("{}: {}: too many retries!".format(retry_function_name, func.__name__))
        raise


def _define_sleep_time(
    sleeptime_kwargs: Optional[Dict[str, Any]], sleeptime_callback: Callable[..., int], attempt: int, func: Callable[..., Any], retry_function_name: str
) -> float:
    sleeptime_kwargs = sleeptime_kwargs or {}
    sleep_time = sleeptime_callback(attempt, **sleeptime_kwargs)
    log.debug("{}: {}: sleeping {} seconds before retry".format(retry_function_name, func.__name__, sleep_time))
    return sleep_time


def retry_async_decorator(
    retry_exceptions: Union[Type[BaseException], Tuple[Type[BaseException], ...]] = Exception, sleeptime_kwargs: Optional[Dict[str, Any]] = None
) -> Callable[..., Callable[..., Awaitable[Any]]]:
    """Decorate a function by wrapping ``retry_async`` around.

    Args:
        retry_exceptions (list or exception, optional): the exception(s) to retry on.
            Defaults to ``Exception``.
        sleeptime_kwargs (dict, optional): the kwargs to pass to ``sleeptime_callback``.
            If None, use {}.  Defaults to None.

    Returns:
        function: the decorated function

    """

    # Better type hinting here once https://www.python.org/dev/peps/pep-0612/ is implemented
    def wrap(async_func: Callable[..., Awaitable[Any]]) -> Callable[..., Awaitable[Any]]:
        @functools.wraps(async_func)
        async def wrapped(*args: Any, **kwargs: Any) -> Any:
            return await retry_async(async_func, retry_exceptions=retry_exceptions, args=args, kwargs=kwargs, sleeptime_kwargs=sleeptime_kwargs)

        return wrapped

    return wrap


# create_temp_creds {{{1
def create_temp_creds(client_id, access_token, start=None, expires=None, scopes=None, name=None):
    """Request temp TC creds with our permanent creds.

    Args:
        client_id (str): the taskcluster client_id to use
        access_token (str): the taskcluster access_token to use
        start (str, optional): the datetime string when the credentials will
            start to be valid.  Defaults to 10 minutes ago, for clock skew.
        expires (str, optional): the datetime string when the credentials will
            expire.  Defaults to 31 days after 10 minutes ago.
        scopes (list, optional): The list of scopes to request for the temp
            creds.  Defaults to ['assume:project:taskcluster:worker-test-scopes', ]
        name (str, optional): the name to associate with the creds.

    Returns:
        dict: the temporary taskcluster credentials.

    """
    now = arrow.utcnow().shift(minutes=-10)
    start = start or now.datetime
    expires = expires or now.shift(days=31).datetime
    scopes = scopes or ["assume:project:taskcluster:worker-test-scopes"]
    creds = createTemporaryCredentials(client_id, access_token, start, expires, scopes, name=name)
    for key, value in creds.items():
        try:
            creds[key] = value.decode("utf-8")
        except (AttributeError, UnicodeDecodeError):
            pass
    return creds


# raise_future_exceptions {{{1
async def raise_future_exceptions(tasks):
    """Given a list of futures, await them, then raise their exceptions if any.

    Without something like this, a bare::

        await asyncio.wait(tasks)

    will swallow exceptions.

    Args:
        tasks (list): the list of futures to await and check for exceptions.

    Returns:
        list: the list of results from the futures.

    Raises:
        Exception: any exceptions in task.exception()

    """
    succeeded_results, _ = await _process_future_exceptions(tasks, raise_at_first_error=True)
    return succeeded_results


# get_results_and_future_exceptions {{{1
async def get_results_and_future_exceptions(tasks):
    """Given a list of futures, await them, then return results and exceptions.

    This is similar to raise_future_exceptions, except that it doesn't raise any
    exception. They are returned instead. This allows some tasks to optionally fail.
    Please consider that no exception will be raised when calling this function.
    You must verify the content of the second item of the tuple. It contains all
    exceptions raised by the futures.

    Args:
        tasks (list): the list of futures to await and check for exceptions.

    Returns:
        tuple: the list of results from the futures, then the list of exceptions.

    """
    return await _process_future_exceptions(tasks, raise_at_first_error=False)


async def _process_future_exceptions(tasks, raise_at_first_error):
    succeeded_results = []
    error_results = []

    if tasks:
        await asyncio.wait(tasks)
        for task in tasks:
            exc = task.exception()
            if exc:
                if raise_at_first_error:
                    raise exc
                else:
                    log.warning("Async task failed with error: {}".format(exc))
                    error_results.append(exc)
            else:
                succeeded_results.append(task.result())

    return succeeded_results, error_results


# filepaths_in_dir {{{1
def filepaths_in_dir(path):
    """Find all files in a directory, and return the relative paths to those files.

    Args:
        path (str): the directory path to walk

    Returns:
        list: the list of relative paths to all files inside of ``path`` or its
            subdirectories.

    """
    filepaths = []
    for root, directories, filenames in os.walk(path):
        for filename in filenames:
            filepath = os.path.join(root, filename)
            filepath = filepath.replace(path, "").lstrip("/")
            filepaths.append(filepath)
    return filepaths


# get_hash {{{1
def get_hash(path, hash_alg="sha256"):
    """Get the hash of the file at ``path``.

    I'd love to make this async, but evidently file i/o is always ready

    Args:
        path (str): the path to the file to hash.
        hash_alg (str, optional): the algorithm to use.  Defaults to 'sha256'.

    Returns:
        str: the hexdigest of the hash.

    """
    h = hashlib.new(hash_alg)
    with open(path, "rb") as f:
        for chunk in iter(functools.partial(f.read, 4096), b""):
            h.update(chunk)
    return h.hexdigest()


# format_json {{{1
def format_json(data):
    """Format json as a sorted string (indents of 2).

    Args:
        data (dict): the json to format.

    Returns:
        str: the formatted json.

    """
    return json.dumps(data, indent=2, sort_keys=True)


# load_json_or_yaml {{{1

# The overload lets us say that exception=None may return None, but if exception kwarg
# is omitted we don't actually ever return None (because on failure we raise an Exception)
@overload
def load_json_or_yaml(
    string: str, is_path: Optional[bool] = ..., file_type: Optional[str] = ..., exception: Type[BaseException] = ..., message: str = ...
) -> Dict[str, Any]:  # pragma: no cover
    ...


@overload
def load_json_or_yaml(
    string: str,
    is_path: Optional[bool] = False,
    file_type: Optional[str] = "json",
    exception: Optional[Type[BaseException]] = ScriptWorkerTaskException,
    message: str = "Failed to load %(file_type)s: %(exc)s",
) -> Optional[Dict[str, Any]]:  # pragma: no cover
    ...


def load_json_or_yaml(
    string: str,
    is_path: Optional[bool] = False,
    file_type: Optional[str] = "json",
    exception: Optional[Type[BaseException]] = ScriptWorkerTaskException,
    message: str = "Failed to load %(file_type)s: %(exc)s",
) -> Optional[Dict[str, Any]]:
    """Load json or yaml from a filehandle or string, and raise a custom exception on failure.

    Args:
        string (str): json/yaml body or a path to open
        is_path (bool, optional): if ``string`` is a path. Defaults to False.
        file_type (str, optional): either "json" or "yaml". Defaults to "json".
        exception (exception, optional): the exception to raise on failure.
            If None, don't raise an exception.  Defaults to ScriptWorkerTaskException.
        message (str, optional): the message to use for the exception.
            Defaults to "Failed to load %(file_type)s: %(exc)s"

    Returns:
        dict: the data from the string.

    Raises:
        Exception: as specified, on failure

    """
    if file_type == "json":
        _load_fh = json.load  # type: Callable[[IO[str]], Dict[str, Any]]
        _load_str = json.loads  # type: Callable[[str], Dict[str, Any]]
    else:
        _load_fh = yaml.safe_load
        _load_str = yaml.safe_load

    try:
        if is_path:
            with open(string, "r") as fh:
                contents = _load_fh(fh)
        else:
            contents = _load_str(string)
        return contents
    except (OSError, ValueError, yaml.scanner.ScannerError) as exc:
        if exception is not None:
            repl_dict = {"exc": str(exc), "file_type": file_type}
            raise exception(message % repl_dict)
    return None


# write_to_file {{{1
def write_to_file(path, contents, file_type="text"):
    """Write ``contents`` to ``path`` with optional formatting.

    Small helper function to write ``contents`` to ``file`` with optional formatting.

    Args:
        path (str): the path to write to
        contents (str, object, or bytes): the contents to write to the file
        file_type (str, optional): the type of file. Currently accepts
            ``text`` or ``binary`` (contents are unchanged) or ``json`` (contents
            are formatted). Defaults to ``text``.

    Raises:
        ScriptWorkerException: with an unknown ``file_type``
        TypeError: if ``file_type`` is ``json`` and ``contents`` isn't JSON serializable

    """
    FILE_TYPES = ("json", "text", "binary")
    if file_type not in FILE_TYPES:
        raise ScriptWorkerException("Unknown file_type {} not in {}!".format(file_type, FILE_TYPES))
    if file_type == "json":
        contents = format_json(contents)
    if file_type == "binary":
        with open(path, "wb") as fh:
            fh.write(contents)
    else:
        with open(path, "w") as fh:
            print(contents, file=fh, end="")


# read_from_file {{{1
def read_from_file(path, file_type="text", exception=ScriptWorkerException):
    """Read from ``path``.

    Small helper function to read from ``file``.

    Args:
        path (str): the path to read from.
        file_type (str, optional): the type of file. Currently accepts
            ``text`` or ``binary``. Defaults to ``text``.
        exception (Exception, optional): the exception to raise
            if unable to read from the file.  Defaults to ``ScriptWorkerException``.

    Returns:
        None: if unable to read from ``path`` and ``exception`` is ``None``
        str or bytes: the contents of ``path``

    Raises:
        Exception: if ``exception`` is set.

    """
    FILE_TYPE_MAP = {"text": "r", "binary": "rb"}
    if file_type not in FILE_TYPE_MAP:
        raise exception("Unknown file_type {} not in {}!".format(file_type, FILE_TYPE_MAP))
    try:
        with open(path, FILE_TYPE_MAP[file_type]) as fh:
            return fh.read()
    except (OSError, FileNotFoundError) as exc:
        raise exception("Can't read_from_file {}: {}".format(path, str(exc)))


# download_file {{{1
async def _log_download_error(resp, msg):
    log.debug(msg, {"url": get_loggable_url(str(resp.url)), "status": resp.status, "body": (await resp.text())[:1000]})
    for i, h in enumerate(resp.history):
        log.debug("Redirect history %s: %s; body=%s", get_loggable_url(str(h.url)), h.status, (await h.text())[:1000])


async def download_file(context, url, abs_filename, session=None, chunk_size=128, auth=None):
    """Download a file, async.

    Args:
        context (scriptworker.context.Context): the scriptworker context.
        url (str): the url to download
        abs_filename (str): the path to download to
        session (aiohttp.ClientSession, optional): the session to use.  If
            None, use context.session.  Defaults to None.
        chunk_size (int, optional): the chunk size to read from the response
            at a time.  Default is 128.

    """
    session = session or context.session
    loggable_url = get_loggable_url(url)
    if auth:
        log.info("Downloading with Authentication %s", loggable_url)
    else:
        log.info("Downloading %s", loggable_url)
    parent_dir = os.path.dirname(abs_filename)
    async with session.get(url, auth=auth) as resp:
        if resp.status == 404:
            await _log_download_error(resp, "404 downloading %(url)s: %(status)s; body=%(body)s")
            raise Download404("{} status {}!".format(loggable_url, resp.status))
        elif resp.status != 200:
            await _log_download_error(resp, "Failed to download %(url)s: %(status)s; body=%(body)s")
            raise DownloadError("{} status {} is not 200!".format(loggable_url, resp.status))
        makedirs(parent_dir)
        with open(abs_filename, "wb") as fd:
            while True:
                chunk = await resp.content.read(chunk_size)
                if not chunk:
                    break
                fd.write(chunk)
    log.info("Done")


# get_loggable_url {{{1
def get_loggable_url(url):
    """Strip out secrets from taskcluster urls.

    Args:
        url (str): the url to strip

    Returns:
        str: the loggable url

    """
    loggable_url = url or ""
    for secret_string in ("bewit=", "AWSAccessKeyId=", "access_token="):
        parts = loggable_url.split(secret_string)
        loggable_url = parts[0]
    if loggable_url != url:
        loggable_url = "{}<snip>".format(loggable_url)
    return loggable_url


def get_parts_of_url_path(url):
    """Given a url, take out the path part and split it by '/'.

    Args:
        url (str): the url slice

    returns
        list: parts after the domain name of the URL

    """
    parsed = urlparse(url)
    path = unquote(parsed.path).lstrip("/")
    parts = path.split("/")
    return parts


# load_json_or_yaml_from_url {{{1
async def load_json_or_yaml_from_url(context: Context, url: str, path: str, overwrite: bool = True, auth: Optional[str] = None) -> Dict[str, Any]:
    """Retry a json/yaml file download, load it, then return its data.

    Args:
        context (scriptworker.context.Context): the scriptworker context.
        url (str): the url to download
        path (str): the path to download to
        overwrite (bool, optional): if False and path exists, don't download.
            Defaults to True.

    Returns:
        dict: the url data.

    Raises:
        Exception: as specified, on failure

    """
    if path.endswith("json"):
        file_type = "json"
    else:
        file_type = "yaml"

    kwargs = {}
    if auth:
        kwargs = {"auth": auth}
    if not overwrite or not os.path.exists(path):
        await retry_async(download_file, args=(context, url, path), kwargs=kwargs, retry_exceptions=(DownloadError, aiohttp.ClientError, asyncio.TimeoutError))
    return load_json_or_yaml(path, is_path=True, file_type=file_type)


# match_url_path_callback {{{1
def match_url_path_callback(match: Match[str]) -> str:
    """Return the path, as a ``match_url_regex`` callback.

    Args:
        match (re.match): the regex match object from ``match_url_regex``

    Returns:
        string: the path matched in the regex.

    """
    path_info = match.groupdict()
    return path_info["path"]


# match_url_regex {{{1
def match_url_regex(rules: Tuple[Any], url: str, callback: Callable[[Match[str]], Any]) -> Any:
    """Given rules and a callback, find the rule that matches the url.

    Rules look like::

        (
            {
                'schemes': ['https', 'ssh'],
                'netlocs': ['hg.mozilla.org'],
                'path_regexes': [
                    "^(?P<path>/mozilla-(central|unified))(/|$)",
                ]
            },
            ...
        )

    Args:
        rules (list): a list of dictionaries specifying lists of ``schemes``,
            ``netlocs``, and ``path_regexes``.
        url (str): the url to test
        callback (function): a callback that takes an ``re.MatchObject``.
            If it returns None, continue searching.  Otherwise, return the
            value from the callback.

    Returns:
        value: the value from the callback, or None if no match.

    """
    parts = urlparse(url)
    path = unquote(parts.path)
    for rule in rules:
        if parts.scheme not in rule["schemes"]:
            continue
        if parts.netloc not in rule["netlocs"]:
            continue
        for regex in rule["path_regexes"]:
            m = re.search(regex, path)
            if m is None:
                continue
            result = callback(m)
            if result is not None:
                return result


# add_enumerable_item_to_dict {{{1
def add_enumerable_item_to_dict(dict_, key, item):
    """Add an item to a list contained in a dict.

    For example: If the dict is ``{'some_key': ['an_item']}``, then calling this function
    will alter the dict to ``{'some_key': ['an_item', 'another_item']}``.

    If the key doesn't exist yet, the function initializes it with a list containing the
    item.

    List-like items are allowed. In this case, the existing list will be extended.

    Args:
        dict_ (dict): the dict to modify
        key (str): the key to add the item to
        item (whatever): The item to add to the list associated to the key

    """
    dict_.setdefault(key, [])
    if isinstance(item, (list, tuple)):
        dict_[key].extend(item)
    else:
        dict_[key].append(item)


# remove_empty_keys {{{1
def remove_empty_keys(values, remove=({}, None, [], "null")):
    """Recursively remove key/value pairs where the value is in ``remove``.

    This is targeted at comparing json-e rebuilt task definitions, since
    json-e drops key/value pairs with empty values.

    Args:
        values (dict/list): the dict or list to remove empty keys from.

    Returns:
        values (dict/list): a dict or list copy, with empty keys removed.

    """
    if isinstance(values, dict):
        return {key: remove_empty_keys(value, remove=remove) for key, value in deepcopy(values).items() if value not in remove}
    if isinstance(values, list):
        return [remove_empty_keys(value, remove=remove) for value in deepcopy(values) if value not in remove]

    return values


# add_projectid {{{1
def add_projectid(task_def):
    """Add a projectId property to a task, if none already exists, using
    the Taskcluster default value of 'none'.

    Args:
        task_def (dict): the task definition

    Returns:
        task_def (dict): the task definition, with projectId
    """
    return {"projectId": "none", **task_def}


# add_taskqueueid {{{1
def add_taskqueueid(task_def):
    """Add a taskQueueId property to a task, if none already exists, based
    on the provisionerId and workerType.  Then remove those two properties.

    Args:
        task_def (dict): the task definition

    Returns:
        task_def (dict): the task definition, with taskQueueId
    """
    if "taskQueueId" not in task_def or "provisionerId" in task_def or "workerType" in task_def:
        task_def = task_def.copy()
    if "taskQueueId" not in task_def:
        task_def["taskQueueId"] = task_def["provisionerId"] + "/" + task_def["workerType"]
    if "provisionerId" in task_def:
        del task_def["provisionerId"]
    if "workerType" in task_def:
        del task_def["workerType"]
    return task_def


# get_single_item_from_sequence {{{1
def get_single_item_from_sequence(
    sequence,
    condition,
    ErrorClass=ValueError,
    no_item_error_message="No item matched condition",
    too_many_item_error_message="Too many items matched condition",
    append_sequence_to_error_message=True,
):
    """Return an item from a python sequence based on the given condition.

    Args:
        sequence (sequence): The sequence to filter
        condition: A function that serves to filter items from `sequence`. Function
            must have one argument (a single item from the sequence) and return a boolean.
        ErrorClass (Exception): The error type raised in case the item isn't unique
        no_item_error_message (str): The message raised when no item matched the condtion
        too_many_item_error_message (str): The message raised when more than one item matched the condition
        append_sequence_to_error_message (bool): Show or hide what was the tested sequence in the error message.
            Hiding it may prevent sensitive data (such as password) to be exposed to public logs

    Returns:
        The only item in the sequence which matched the condition

    """
    filtered_sequence = [item for item in sequence if condition(item)]
    number_of_items_in_filtered_sequence = len(filtered_sequence)
    if number_of_items_in_filtered_sequence == 0:
        error_message = no_item_error_message
    elif number_of_items_in_filtered_sequence > 1:
        error_message = too_many_item_error_message
    else:
        return filtered_sequence[0]

    if append_sequence_to_error_message:
        error_message = "{}. Given: {}".format(error_message, sequence)
    raise ErrorClass(error_message)


# semaphore_wrapper {{{1
async def semaphore_wrapper(semaphore, coro):
    """Wrap an async function with semaphores.

    Usage::

        semaphore = asyncio.Semaphore(10)  # max 10 concurrent
        futures = []
        futures.append(asyncio.ensure_future(semaphore_wrapper(
            semaphore, do_something(arg1, arg2, kwarg1='foo')
        )))
        await raise_future_exceptions(futures)

    Args:
        semaphore (asyncio.Semaphore): the semaphore to wrap the action with
        coro (coroutine): an asyncio coroutine

    Returns:
        the result of ``action``.

    """
    async with semaphore:
        return await coro



#!/usr/bin/env python
"""scriptworker context.

Most functions need access to a similar set of objects.  Rather than
having to pass them all around individually or create a monolithic 'self'
object, let's point to them from a single context object.

Attributes:
    log (logging.Logger): the log object for the module.
    DEFAULT_MAX_CONCURRENT_DOWNLOADS (int): default max concurrent downloads

"""
import asyncio
import json
import logging
import os
import tempfile
from copy import deepcopy
from typing import Any, Dict, Optional, cast

import aiohttp
import arrow
from taskcluster.aio import Queue

from scriptworker import task_process
from scriptworker.exceptions import CoTError
from scriptworker.utils import load_json_or_yaml_from_url, makedirs

log = logging.getLogger(__name__)


DEFAULT_MAX_CONCURRENT_DOWNLOADS = 5


class Context(object):
    """Basic config holding object.

    Avoids putting everything in single monolithic object, but allows for
    passing around config and easier overriding in tests.

    Attributes:
        config (dict): the running config.  In production this will be an
            immutabledict.
        credentials_timestamp (int): the unix timestamp when we last updated
            our credentials.
        proc (task_process.TaskProcess): when launching the script, this is
            the process object.
        queue (taskcluster.aio.Queue): the taskcluster Queue object
            containing the scriptworker credentials.
        session (aiohttp.ClientSession): the default aiohttp session
        task (dict): the task definition for the current task.
        temp_queue (taskcluster.aio.Queue): the taskcluster Queue object
            containing the task-specific temporary credentials.

    """

    config: Optional[Dict[str, Any]] = None
    credentials_timestamp: Optional[int] = None
    proc: Optional[task_process.TaskProcess] = None
    queue: Optional[Queue] = None
    session: Optional[aiohttp.ClientSession] = None
    task: Optional[Dict[str, Any]] = None
    temp_queue: Optional[Queue] = None
    running_tasks = None
    _download_semaphore = None
    _credentials: Optional[Dict[str, Any]] = None
    _claim_task: Optional[Dict[str, Any]] = None  # This assumes a single task per worker.
    _event_loop = None
    _temp_credentials = None  # This assumes a single task per worker.
    _reclaim_task = None
    _projects = None

    @property
    def claim_task(self) -> Optional[Dict[str, Any]]:
        """dict: The current or most recent claimTask definition json from the queue.

        This contains the task definition, as well as other task-specific
        info.

        When setting ``claim_task``, we also set ``self.task`` and
        ``self.temp_credentials``, zero out ``self.reclaim_task`` and ``self.proc``,
        then write a task.json to disk.

        """
        return self._claim_task

    @claim_task.setter
    def claim_task(self, claim_task: Optional[Dict[str, Any]]) -> None:
        assert self.config
        self._claim_task = claim_task
        self.reclaim_task = None
        self.proc = None
        if claim_task:
            self.task = claim_task["task"]
            self.verify_task()
            self.temp_credentials = claim_task["credentials"]
            path = os.path.join(self.config["work_dir"], "task.json")
            assert self.task
            self.write_json(path, self.task, "Writing task file to {path}...")
        else:
            self.temp_credentials = None
            self.task = None

    def verify_task(self) -> None:
        """Run some task sanity checks on ``self.task``."""
        assert self.task
        for upstream_artifact in self.task.get("payload", {}).get("upstreamArtifacts", []):
            task_id: str = upstream_artifact["taskId"]
            for path in upstream_artifact["paths"]:
                if os.path.isabs(path) or ".." in path:
                    raise CoTError("upstreamArtifacts taskId {} has illegal path {}!".format(task_id, path))

    @property
    def credentials(self) -> Optional[Dict[str, Any]]:
        """dict: The current scriptworker credentials.

        These come from the config or CREDS_FILES or environment.

        When setting credentials, also create a new ``self.queue`` and
        update self.credentials_timestamp.

        """
        if self._credentials:
            return dict(deepcopy(self._credentials))
        return None

    @credentials.setter
    def credentials(self, creds: Optional[Dict[str, Any]]) -> None:
        self._credentials = creds
        self.queue = self.create_queue(self.credentials)
        self.credentials_timestamp = arrow.utcnow().int_timestamp

    @property
    def task_id(self) -> Optional[str]:
        """string: The running task's taskId."""
        if self.claim_task:
            return cast(str, self.claim_task["status"]["taskId"])
        return None

    def create_queue(self, credentials: Optional[Dict[str, Any]]) -> Optional[Queue]:
        """Create a taskcluster queue.

        Args:
            credentials (dict): taskcluster credentials.

        """
        assert self.config
        if credentials:
            session = self.session or aiohttp.ClientSession(loop=self.event_loop)
            return Queue(options={"credentials": credentials, "rootUrl": self.config["taskcluster_root_url"]}, session=session)
        return None

    @property
    def reclaim_task(self) -> Optional[Dict[str, Any]]:
        """dict: The most recent reclaimTask definition.

        This contains the newest expiration time and the newest temp credentials.

        When setting ``reclaim_task``, we also set ``self.temp_credentials``.

        ``reclaim_task`` will be ``None`` if there hasn't been a claimed task yet,
        or if a task has been claimed more recently than the most recent
        reclaimTask call.

        """
        return self._reclaim_task

    @reclaim_task.setter
    def reclaim_task(self, value: Optional[Dict[str, Any]]) -> None:
        self._reclaim_task = value
        if value is not None:
            self.temp_credentials = value["credentials"]

    @property
    def temp_credentials(self) -> Optional[Dict[str, Any]]:
        """dict: The latest temp credentials, or None if we haven't claimed a task yet.

        When setting, create ``self.temp_queue`` from the temp taskcluster creds.

        """
        if self._temp_credentials:
            return dict(deepcopy(self._temp_credentials))
        return None

    @temp_credentials.setter
    def temp_credentials(self, credentials: Optional[Dict[str, Any]]) -> None:
        self._temp_credentials = credentials
        self.temp_queue = self.create_queue(self.temp_credentials)

    def write_json(self, path: str, contents: Dict[str, Any], message: str) -> None:
        """Write json to disk.

        Args:
            path (str): the path to write to
            contents (dict): the contents of the json blob
            message (str): the message to log

        """
        log.debug(message.format(path=path))
        makedirs(os.path.dirname(path))
        with open(path, "w") as fh:
            json.dump(contents, fh, indent=2, sort_keys=True)

    @property
    def projects(self) -> Optional[Dict[str, Any]]:
        """dict: The current contents of ``projects.yml``, which defines CI configuration.

        I'd love to auto-populate this; currently we need to set this from
        the config's ``project_configuration_url``.

        """
        if self._projects:
            return dict(deepcopy(self._projects))
        return None

    @projects.setter
    def projects(self, projects: Optional[Dict[str, Any]]) -> None:
        self._projects = projects

    @property
    def event_loop(self) -> asyncio.AbstractEventLoop:
        """asyncio.BaseEventLoop: the running event loop.

        This fixture mainly exists to allow for overrides during unit tests.

        """
        if not self._event_loop:
            self._event_loop = asyncio.get_event_loop()
        return self._event_loop

    @event_loop.setter
    def event_loop(self, event_loop: asyncio.AbstractEventLoop) -> None:
        self._event_loop = event_loop

    async def populate_projects(self, force: bool = False) -> None:
        """Download the ``projects.yml`` file and populate ``self.projects``.

        This only sets it once, unless ``force`` is set.

        Args:
            force (bool, optional): Re-run the download, even if ``self.projects``
                is already defined. Defaults to False.

        """
        assert self.config
        if force or not self.projects:
            with tempfile.TemporaryDirectory() as tmpdirname:
                self.projects = await load_json_or_yaml_from_url(self, self.config["project_configuration_url"], os.path.join(tmpdirname, "projects.yml"))

    @property
    def download_semaphore(self) -> asyncio.BoundedSemaphore:
        assert self.config
        if self._download_semaphore is None:
            try:
                max_concurrent_downloads = self.config.get("max_concurrent_downloads", DEFAULT_MAX_CONCURRENT_DOWNLOADS)
            except (TypeError, KeyError, AttributeError):
                max_concurrent_downloads = DEFAULT_MAX_CONCURRENT_DOWNLOADS
            self._download_semaphore = asyncio.BoundedSemaphore(max_concurrent_downloads)
        return self._download_semaphore



#!/usr/bin/env python
"""Scriptworker worker functions.

Attributes:
    log (logging.Logger): the log object for the module.

"""
import asyncio
import logging
import os
import signal
import socket
import sys
import typing
from typing import Any

import aiohttp
import arrow

from scriptworker.artifacts import upload_artifacts
from scriptworker.config import get_context_from_cmdln
from scriptworker.constants import STATUSES
from scriptworker.cot.generate import generate_cot
from scriptworker.cot.verify import ChainOfTrust, verify_chain_of_trust
from scriptworker.exceptions import ScriptWorkerException, WorkerShutdownDuringTask
from scriptworker.task import claim_work, complete_task, prepare_to_run_task, reclaim_task, run_task, worst_level
from scriptworker.task_process import TaskProcess
from scriptworker.utils import cleanup, filepaths_in_dir

log = logging.getLogger(__name__)


# do_run_task {{{1
async def do_run_task(context, run_cancellable, to_cancellable_process):
    """Run the task logic.

    Returns the integer status of the task.

    args:
        context (scriptworker.context.Context): the scriptworker context.
        run_cancellable (typing.Callable): wraps future such that it'll cancel upon worker shutdown
        to_cancellable_process (typing.Callable): wraps ``TaskProcess`` such that it will stop if the worker is shutting
            down

    Raises:
        Exception: on unexpected exception.

    Returns:
        int: exit status

    """
    status = 0
    try:
        if context.config["verify_chain_of_trust"]:
            chain = ChainOfTrust(context, context.config["cot_job_type"])
            await run_cancellable(verify_chain_of_trust(chain))
        status = await run_task(context, to_cancellable_process)
        generate_cot(context)
    except asyncio.CancelledError:
        log.info("CoT cancelled asynchronously")
        raise WorkerShutdownDuringTask
    except ScriptWorkerException as e:
        status = worst_level(status, e.exit_code)
        log.error("Hit ScriptWorkerException: {}".format(e))
    except Exception as e:
        log.exception("SCRIPTWORKER_UNEXPECTED_EXCEPTION task {}".format(e))
        status = STATUSES["internal-error"]
    return status


# do_upload {{{1
async def do_upload(context, files):
    """Upload artifacts and return status.

    Returns the integer status of the upload.

    args:
        context (scriptworker.context.Context): the scriptworker context.
        files (list of str): list of files to be uploaded as artifacts

    Raises:
        Exception: on unexpected exception.

    Returns:
        int: exit status

    """
    status = 0
    try:
        await upload_artifacts(context, files)
    except ScriptWorkerException as e:
        status = worst_level(status, e.exit_code)
        log.error("Hit ScriptWorkerException: {}".format(e))
    except (aiohttp.ClientError, asyncio.TimeoutError) as e:
        status = worst_level(status, STATUSES["intermittent-task"])
        log.error("Hit {}: {}".format(type(e), e))
    except Exception as e:
        log.exception("SCRIPTWORKER_UNEXPECTED_EXCEPTION upload {}".format(e))
        raise
    return status


class RunTasks:
    """Manages processing of Taskcluster tasks."""

    def __init__(self):
        """Constructor."""
        self.future = None
        self.task_process = None
        self.is_cancelled = False

    async def invoke(self, context):
        """Claims and processes Taskcluster work.

        Args:
            context (scriptworker.context.Context): context of worker

        Returns: status code of build

        """
        try:
            # Note: claim_work(...) might not be safely interruptible! See
            # https://bugzilla.mozilla.org/show_bug.cgi?id=1524069
            tasks = await self._run_cancellable(claim_work(context))
            if not tasks or not tasks.get("tasks", []):
                await self._run_cancellable(asyncio.sleep(context.config["poll_interval"]))
                return None

            # Assume only a single task, but should more than one fall through,
            # run them sequentially.  A side effect is our return status will
            # be the status of the final task run.
            status = None
            for task_defn in tasks.get("tasks", []):
                prepare_to_run_task(context, task_defn)
                reclaim_fut = context.event_loop.create_task(reclaim_task(context, context.task))
                try:
                    status = await do_run_task(context, self._run_cancellable, self._to_cancellable_process)
                    artifacts_paths = filepaths_in_dir(context.config["artifact_dir"])
                except WorkerShutdownDuringTask:
                    shutdown_artifact_paths = [os.path.join("public", "logs", log_file) for log_file in ["chain_of_trust.log", "live_backing.log"]]
                    artifacts_paths = [path for path in shutdown_artifact_paths if os.path.isfile(os.path.join(context.config["artifact_dir"], path))]
                    status = STATUSES["worker-shutdown"]
                status = worst_level(status, await do_upload(context, artifacts_paths))
                await complete_task(context, status)
                reclaim_fut.cancel()
                cleanup(context)

            return status

        except asyncio.CancelledError:
            return None

    async def _run_cancellable(self, coroutine: typing.Awaitable[Any]) -> Any:
        self.future = asyncio.ensure_future(coroutine)
        if self.is_cancelled:
            self.future.cancel()
        result = await self.future
        self.future = None
        return result

    async def _to_cancellable_process(self, task_process: TaskProcess) -> TaskProcess:
        self.task_process = task_process

        if self.is_cancelled:
            await task_process.worker_shutdown_stop()

        return task_process

    async def cancel(self):
        """Cancel current work."""
        self.is_cancelled = True
        if self.future is not None:
            self.future.cancel()
        if self.task_process is not None:
            log.warning("Worker is shutting down, but a task is running. Terminating task")
            await self.task_process.worker_shutdown_stop()


# run_tasks {{{1
async def run_tasks(context):
    """Run any tasks returned by claimWork.

    Returns the integer status of the task that was run, or None if no task was
    run.

    args:
        context (scriptworker.context.Context): the scriptworker context.

    Raises:
        Exception: on unexpected exception.

    Returns:
        int: exit status
        None: if no task run.

    """
    running_tasks = RunTasks()
    context.running_tasks = running_tasks
    status = await running_tasks.invoke(context)
    context.running_tasks = None
    return status


# async_main {{{1
async def async_main(context, credentials):
    """Set up and run tasks for this iteration.

    https://firefox-ci-tc.services.mozilla.com/docs/reference/platform/queue/worker-interaction

    Args:
        context (scriptworker.context.Context): the scriptworker context.
    """
    async with aiohttp.ClientSession() as session:
        context.session = session
        context.credentials = credentials
        await run_tasks(context)


# main {{{1
def main(event_loop=None):
    """Scriptworker entry point: get everything set up, then enter the main loop.

    Args:
        event_loop (asyncio.BaseEventLoop, optional): the event loop to use.
            If None, use ``asyncio.get_event_loop()``. Defaults to None.

    """
    context, credentials = get_context_from_cmdln(sys.argv[1:])
    log.info("Scriptworker starting up at {} UTC".format(arrow.utcnow().format()))
    log.info("Worker FQDN: {}".format(socket.getfqdn()))
    cleanup(context)
    context.event_loop = event_loop or asyncio.get_event_loop()

    done = False

    async def _handle_sigterm():
        log.info("SIGTERM received; shutting down")
        nonlocal done
        done = True
        if context.running_tasks is not None:
            await context.running_tasks.cancel()

    async def _handle_sigusr1():
        """Stop accepting new tasks."""
        log.info("SIGUSR1 received; no more tasks will be taken")
        nonlocal done
        done = True

    context.event_loop.add_signal_handler(signal.SIGTERM, lambda: asyncio.ensure_future(_handle_sigterm()))
    context.event_loop.add_signal_handler(signal.SIGUSR1, lambda: asyncio.ensure_future(_handle_sigusr1()))

    while not done:
        try:
            context.event_loop.run_until_complete(async_main(context, credentials))
        except Exception:
            log.critical("Fatal exception", exc_info=1)
            raise
    else:
        log.info("Scriptworker stopped at {} UTC".format(arrow.utcnow().format()))
        log.info("Worker FQDN: {}".format(socket.getfqdn()))



#!/usr/bin/env python
"""Generate an ed25519 keypair, and store as base64-encoded text files.

This script doesn't currently reuse the functions in `scriptworker.ed25519`, for
easier standalone use. It could easily be a `console_script` though.

"""
from __future__ import print_function

import base64
import sys

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey, Ed25519PublicKey


def private_key_from_string(key_str):
    """Create an Ed25519PrivateKey from a base64-encoded string."""
    return Ed25519PrivateKey.from_private_bytes(base64.b64decode(key_str))


def public_key_from_string(key_str):
    """Create an Ed25519PublicKey from a base64-encoded string."""
    return Ed25519PublicKey.from_public_bytes(base64.b64decode(key_str))


def b64_from_private_key(key):
    """Get the base64 string from an Ed25519PrivateKey."""
    return base64.b64encode(
        key.private_bytes(encoding=serialization.Encoding.Raw, format=serialization.PrivateFormat.Raw, encryption_algorithm=serialization.NoEncryption())
    ).decode("utf-8")


def b64_from_public_key(key):
    """Get the base64 string from an Ed25519PublicKey."""
    return base64.b64encode(key.public_bytes(encoding=serialization.Encoding.Raw, format=serialization.PublicFormat.Raw)).decode("utf-8")


prefix = ""
if len(sys.argv) > 1:
    prefix = "{}_".format(sys.argv[1])

privkey = Ed25519PrivateKey.generate()
pubkey = privkey.public_key()
privkey_str = b64_from_private_key(privkey)
pubkey_str = b64_from_public_key(pubkey)

# test
privkey2 = private_key_from_string(privkey_str)
pubkey2 = public_key_from_string(pubkey_str)
assert b64_from_private_key(privkey2) == privkey_str
assert b64_from_public_key(pubkey2) == pubkey_str

with open("{}private_key".format(prefix), "w") as fh:
    fh.write(privkey_str)
with open("{}public_key".format(prefix), "w") as fh:
    fh.write(pubkey_str)



#!/usr/bin/env python
"""TaskProcess class to provide helper methods around worker task's process.

Attributes:
    log (logging.Logger): the log object for this module

"""

import asyncio
import logging
import os
import signal
from asyncio.subprocess import Process

log = logging.getLogger(__name__)


class TaskProcess:
    """Wraps worker task's process."""

    def __init__(self, process: Process):
        """Constructor.

        Args:
            process (Process): task process
        """
        self.process = process
        self.stopped_due_to_worker_shutdown = False

    async def worker_shutdown_stop(self) -> None:
        """Invoke on worker shutdown to stop task process."""
        self.stopped_due_to_worker_shutdown = True
        await self.stop()

    async def stop(self) -> None:
        """Stop the current task process.

        Starts with SIGTERM, gives the process 1 second to terminate, then kills it
        """
        # negate pid so that signals apply to process group
        pgid = -self.process.pid
        try:
            os.kill(pgid, signal.SIGTERM)
            await asyncio.sleep(1)
            os.kill(pgid, signal.SIGKILL)
        except (OSError, ProcessLookupError):
            return



#!/usr/bin/env python
"""Scriptworker logging.

Attributes:
    log (logging.Logger): the log object for this module.

"""
import logging
import logging.handlers
import os
from asyncio.streams import StreamReader
from contextlib import contextmanager
from typing import IO, Any, Generator, Iterator, Optional, Sequence, Union

from scriptworker.utils import makedirs, to_unicode

log = logging.getLogger(__name__)


def update_logging_config(context: Any, log_name: Optional[str] = None, file_name: str = "worker.log") -> None:
    """Update python logging settings from config.

    By default, this sets the ``scriptworker`` log settings, but this will
    change if some other package calls this function or specifies the ``log_name``.

    * Use formatting from config settings.
    * Log to screen if ``verbose``
    * Add a rotating logfile from config settings.

    Args:
        context (scriptworker.context.Context): the scriptworker context.
        log_name (str, optional): the name of the Logger to modify.
            If None, use the top level module ('scriptworker').
            Defaults to None.

    """
    log_name = log_name or __name__.split(".")[0]
    top_level_logger = logging.getLogger(log_name)

    datefmt = context.config["log_datefmt"]
    fmt = context.config["log_fmt"]
    formatter = logging.Formatter(fmt=fmt, datefmt=datefmt)

    if context.config.get("verbose"):
        top_level_logger.setLevel(logging.DEBUG)
        if len(top_level_logger.handlers) == 0:
            handler = logging.StreamHandler()
            handler.setFormatter(formatter)
            top_level_logger.addHandler(handler)
    else:
        top_level_logger.setLevel(logging.INFO)

    # Rotating log file
    makedirs(context.config["log_dir"])
    path = os.path.join(context.config["log_dir"], file_name)
    if context.config["watch_log_file"]:
        # If we rotate the log file via logrotate.d, let's watch the file
        # so we can automatically close/reopen on move.
        handler = logging.handlers.WatchedFileHandler(path)  # type: ignore
    elif context.config["log_max_bytes"] and context.config["log_max_backups"]:
        handler = logging.handlers.RotatingFileHandler(  # type: ignore
            filename=path,
            maxBytes=context.config["log_max_bytes"],
            backupCount=context.config["log_max_backups"],
        )
    else:
        # Avoid using WatchedFileHandler during scriptworker unittests
        handler = logging.FileHandler(path)  # type: ignore
    handler.setFormatter(formatter)
    top_level_logger.addHandler(handler)
    top_level_logger.addHandler(logging.NullHandler())


async def pipe_to_log(pipe: StreamReader, filehandles: Sequence[IO[str]] = (), level: int = logging.INFO) -> None:
    """Log from a subprocess PIPE.

    Args:
        pipe (filehandle): subprocess process STDOUT or STDERR
        filehandles (list of filehandles, optional): the filehandle(s) to write
            to.  If empty, don't write to a separate file.  Defaults to ().
        level (int, optional): the level to log to.  Defaults to ``logging.INFO``.

    """
    while True:
        line = await pipe.readline()  # type: Union[str, bytes]
        if line:
            line = to_unicode(line)
            log.log(level, line.rstrip())
            for filehandle in filehandles:
                print(line, file=filehandle, end="")
        else:
            break


def get_log_filename(context: Any) -> str:
    """Get the task log/error file paths.

    Args:
        context (scriptworker.context.Context): the scriptworker context.

    Returns:
        string: log file path

    """
    # XXX Even though our logs aren't live, Treeherder looks for live_backing.log to show errors in failures summary
    return os.path.join(context.config["task_log_dir"], "live_backing.log")


@contextmanager
def get_log_filehandle(context: Any) -> Iterator[IO[str]]:
    """Open the log and error filehandles.

    Args:
        context (scriptworker.context.Context): the scriptworker context.

    Yields:
        log filehandle

    """
    log_file_name = get_log_filename(context)
    makedirs(context.config["task_log_dir"])
    with open(log_file_name, "w", encoding="utf-8") as filehandle:
        yield filehandle


@contextmanager
def contextual_log_handler(
    context: Any, path: str, log_obj: Optional[logging.Logger] = None, level: int = logging.DEBUG, formatter: Optional[logging.Formatter] = None
) -> Generator[None, None, None]:
    """Add a short-lived log with a contextmanager for cleanup.

    Args:
        context (scriptworker.context.Context): the scriptworker context
        path (str): the path to the log file to create
        log_obj (logging.Logger): the log object to modify.  If None, use
            ``scriptworker.log.log``.  Defaults to None.
        level (int, optional): the logging level.  Defaults to logging.DEBUG.
        formatter (logging.Formatter, optional): the logging formatter. If None,
            defaults to ``logging.Formatter(fmt=fmt)``. Default is None.

    Yields:
        None: but cleans up the handler afterwards.

    """
    log_obj = log_obj or log
    formatter = formatter or logging.Formatter(fmt=context.config["log_fmt"], datefmt=context.config["log_datefmt"])
    parent_path = os.path.dirname(path)
    makedirs(parent_path)
    contextual_handler = logging.FileHandler(path, encoding="utf-8")
    contextual_handler.setLevel(level)
    contextual_handler.setFormatter(formatter)
    log_obj.addHandler(contextual_handler)
    yield
    contextual_handler.close()
    log_obj.removeHandler(contextual_handler)



from __future__ import print_function

import json
import os
import sys

from setuptools import setup
from setuptools.command.test import test as TestCommand

if {"register", "upload"}.intersection(set(sys.argv)):
    print(
        "                        ***** WARNING *****\n"
        "`python setup.py register` and `python setup.py upload` are unsafe!\n"
        "See http://scriptworker.readthedocs.io/en/latest/releases.html#pypi\n"
        "\n"
        "Exiting...",
        file=sys.stderr,
    )
    sys.exit(1)

tests_require = [
    "asyncio_extras",
    "flake8",
    "flake8_docstrings",
    "mock",
    "pytest",
    "pytest-asyncio",
    "pytest-cov",
    "pytest-mock",
    "pytest-random-order",
    "tox",
    "virtualenv",
]

PATH = os.path.join(os.path.dirname(__file__), "version.json")
with open(PATH) as filehandle:
    VERSION = json.load(filehandle)["version_string"]

with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), "requirements.txt")) as f:
    install_requires = f.readlines()


class Tox(TestCommand):
    """http://bit.ly/1T0dwvG"""

    user_options = [("tox-args=", "a", "Arguments to pass to tox")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.tox_args = None

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import shlex

        import tox

        args = self.tox_args
        if args:
            args = shlex.split(self.tox_args)
        errno = tox.cmdline(args=args)
        sys.exit(errno)


setup(
    name="scriptworker",
    version=VERSION,
    description="TaskCluster Script Worker",
    author="Mozilla Release Engineering",
    author_email="release+python@mozilla.com",
    url="https://github.com/mozilla-releng/scriptworker",
    packages=["scriptworker", "scriptworker.cot"],
    package_data={"": ["version.json"]},
    package_dir={"": "src"},
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "scriptworker = scriptworker.worker:main",
            "verify_cot = scriptworker.cot.verify:verify_cot_cmdln",
            "create_test_workdir = scriptworker.cot.verify:create_test_workdir",
            "verify_ed25519_signature = scriptworker.ed25519:verify_ed25519_signature_cmdln",
        ]
    },
    zip_safe=False,
    license="MPL 2.0",
    install_requires=install_requires,
    tests_require=tests_require,
    python_requires=">=3.7",
    cmdclass={"test": Tox},
    classifiers=(
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ),
)
